"""
Input encoding
"""
import inspect
import numpy as np
import scipy.signal as signal
import copy

from .utils import get_default_kwargs

"""
Shapes of signals

(time,) -> (time, 1)
1D arrays are 1D signals

(time, D)
2D arrays are D-dimensional signals
With D=2 we have a vector signal

(time, H, W)
3D arrays are signals on a WxH grid

(time, H, W, D)
4D arrays are D-dimensional signals on a WxH grid
With D=2 we have a vector signal on a grid

In general signals can take any shape as part of the encoding process.
However, the last step must produce an output of either
(time, 2): a global vector signal
(time, H, W, 2): a local vector signal on a grid

"""
class Encoder:
    def __init__(self, *steps, verbose=False):
        self.steps = steps
        self.params = [self._get_default_params(s) for s in steps]
        self.verbose = verbose

    def _verbose(self, *args, **kwargs):
        if self.verbose:
            print(*args, **kwargs)

    def encode(self, input, verbose=False):
        """ Encode input stream

        input.shape = (n_inputs, input_dim)
        output.shape = (n_inputs, time, output_dim, ...) # step
                     = (n_inputs, time, 2)               # global field
                     = (n_inputs, time, height, width)   # local field on grid
        """
        input = np.array(input)
        self._verbose(f'input {input.shape}', end=' ')
        for step, params in zip(self.steps, self.params):
            input = step(input, **params)
            self._verbose(f'-> {step.__name__}({params})', end=' ')
            self._verbose(f'-> {input.shape}', end=' ')
        self._verbose('')
        return input

    def __call__(self, input):
        return self.encode(input)

    @staticmethod
    def _get_default_params(step):
        return dict(get_default_kwargs(step))

    def get_default_params(self):
        params = {}
        for p in self.get_params():
            params.update(p)
        return params

    def get_params(self):
        return self.params

    def set_param(self, name, value):
        """ Set all params matching name """
        for param in self.params:
            if name in param:
                param[name] = value

    def set_params(self, **params):
        for k, v in params.items():
            self.set_param(k, v)

    def __repr__(self):
        return f'Encoder({self.steps})'

def expand_dims(input):
    """ Add a new dimension at the end """
    return np.expand_dims(input, -1)

def check_input(input):
    input = np.array(input)
    if len(input.shape) == 1:
        return expand_dims(input)
    return input

def squeeze(input):
    """ Squeeze all but the first axis """
    time = input.shape[0]
    input_shape = input.shape[1:]
    input_shape = tuple(dim for dim in input_shape if dim > 1)
    return input.reshape((time,) + input_shape)

# Steps as simple functions
def scale(input, H0=0, H=1):
    """ Scale input to range [H0, H] """
    return H0 + input * (H - H0)

def angle(input, phi0=0, phi=360):
    """ Map input to angle in range [phi0, phi] -> vector components """
    input = check_input(input)
    theta0 = np.deg2rad(phi0)
    theta = np.deg2rad(phi)
    angles = theta0 + input * (theta - theta0)
    # add one dimension for vector
    out = expand_dims(angles)
    out = np.repeat(out, 2, axis=-1)
    out[...,0] = np.cos(out[...,0])
    out[...,1] = np.sin(out[...,1])
    return squeeze(out)

def angle_grid(input, phi0=0, phi=360, grid=None):
    """ Convenience function to offset all angles based on grid """
    input = expand_dims(check_input(input))
    if grid is not None:
        phi0 = grid + phi0
        phi = phi0 + phi
    return angle(input, phi0, phi)


def sw(input, rotation=0):
    """ Multiply input vectors by SW astroid """
    input = check_input(input)
    angles = np.arctan2(input[...,1], input[...,0])
    angles = angles - np.deg2rad(rotation)

    Hs = (np.abs(np.sin(angles))**(2/3) + np.abs(np.cos(angles))**(2/3))**(-3/2)
    Hs.shape += (1,)

    return Hs * input

def broadcast_waveform(input, waveform):
    """ Broadcast waveform over input """
    input = check_input(input)
    # multiply input by waveform (broadcasting rules apply)
    out = input * waveform
    # roll last axis to the first (time)
    out = np.moveaxis(out, -1, 1)
    # finally concatenate it
    out = np.concatenate(out)
    return squeeze(out)

def sin(input, timesteps=100):
    """ Multiply input by sine wave """
    t = np.linspace(0, 2 * np.pi, timesteps, endpoint=False)
    waveform = np.sin(t)
    return broadcast_waveform(input, waveform)

def triangle(input, timesteps=100):
    """ Multiply input by triangle wave """
    t = np.linspace(np.pi/2, 2*np.pi + np.pi/2, timesteps, endpoint=False)
    waveform = signal.sawtooth(t, 0.5)
    return broadcast_waveform(input, waveform)

def rotate(input, timesteps=100):
    """ Multiply input by rotation vector """
    t = np.linspace(0, 2 * np.pi, timesteps, endpoint=False)
    cos = broadcast_waveform(input, np.cos(t))
    sin = broadcast_waveform(input, np.sin(t))
    return np.stack([cos, sin], axis=-1)

def ppm(input, pos0=0, pos=1.0, timesteps=100):
    """ Pulse position modulation """
    input = check_input(input)

    # output shape
    shape = input.shape + (timesteps,)

    # calculate location of pulses
    pulse_inds = scale(input, pos0 * timesteps, pos * (timesteps-1)).round().astype(int)
    inds = tuple(np.array(list(np.ndindex(shape[:-1]))).T)
    inds += (pulse_inds.flatten(),)

    # make the pulses
    out = signal.unit_impulse(shape, inds)

    return squeeze(out)

def fixed_vector(input, phi=0):
    """ Convert scalar input to vectors at some angle phi """
    input = check_input(input)
    theta = np.deg2rad(phi)
    # add one dimension to broadcast vector
    input = expand_dims(input)
    vector = [np.cos(theta), np.sin(theta)]
    out = input * vector
    return squeeze(out)

def rotate_vector(input, phi=0):
    """ Rotate vector input by some angle phi """
    input = check_input(input)
    input_shape = input.shape
    theta = np.deg2rad(phi)
    rotation_matrix = np.array(
        [[np.cos(theta), -np.sin(theta)],
         [np.sin(theta),  np.cos(theta)]])
    input = input.reshape((-1,2)).T
    out = np.dot(rotation_matrix, input).T
    out = out.reshape(input_shape)
    return squeeze(out)

def ensure2d(input, fill=1.0):
    """ Convert scalar input to vectors """
    input = check_input(input)
    if input.shape[-1] == 1:
        input = np.column_stack([input, np.full(len(input), fill)])
    return input

def grid(input, grid_size=(3,3), grid=None):
    """ Encode input onto grid

    Parameters
    ----------
    grid_size : tuple
        Create grid of fixed size (uniform weights)
        Ignored if grid is set
    grid : array
        Create grid of given size (non-uniform weights)
    """
    if grid is None:
        w, h = grid_size
        grid = np.ones((h, w))
    else:
        grid = np.atleast_2d(grid)

    input = check_input(input)

    # add dimensions to broadcast the grid
    input = input.reshape(input.shape + (1,) * grid.ndim)

    # broadcast the grid on each feature
    out = grid * input

    # now we have:
    # out.shape = (time,) + input.shape + grid.shape
    # shift axes to the end until we have the desired:
    # out.shape = (time,) + grid.shape + input.shape
    for i in range(out.ndim - grid.ndim - 1):
        out = np.moveaxis(out, 1, -1)

    return squeeze(out)

def onehot(input, nbits=2):
    input = check_input(input)

    # roll bit string 00..01 according to input
    out = np.zeros(input.shape + (nbits,))
    out[...,0] = 1
    for i, v in np.ndenumerate(input.astype(int)):
        out[i] = np.roll(out[i], v)

    # roll last axis to the first (time)
    out = np.moveaxis(out, -1, 1)

    # finally concatenate it
    out = np.concatenate(out)

    return squeeze(out)

"""
Define some encoders
"""
encoders = {
    'constant': Encoder(scale, fixed_vector),
    'angle': Encoder(angle, scale),
    'sw-angle': Encoder(angle, sw, scale),
    'sin': Encoder(scale, sin, fixed_vector),
    'triangle': Encoder(scale, triangle, fixed_vector),
    'angle-sin': Encoder(angle, expand_dims, sin, scale),
    'angle-triangle': Encoder(angle, expand_dims, triangle, scale),
    'angle-sin-grid': Encoder(angle, expand_dims, sin, scale, grid), # grid determines H
    'angle-grid-sin': Encoder(angle_grid, expand_dims, sin, scale), # grid adds angle offset
    'rotate': Encoder(scale, rotate),
    'sw-rotate': Encoder(scale, rotate, sw),
    'rotate-pulse': Encoder(ppm, scale, rotate),
    'sw-pulse': Encoder(ppm, scale, rotate, sw),
    'triangle2d': Encoder(scale, ensure2d, triangle, rotate_vector),
    'onehot': Encoder(onehot, scale, triangle, fixed_vector),
    'constant-grid': Encoder(grid, scale, fixed_vector),
    'sin-grid': Encoder(grid, scale, expand_dims, sin, fixed_vector),
    'triangle-grid': Encoder(grid, scale, expand_dims, triangle, fixed_vector),
    'rotate-pulse-grid': Encoder(ppm, grid, scale, rotate),
    'sw-pulse-grid': Encoder(ppm, grid, scale, rotate, sw),
    'direct': Encoder(scale),
}

def get_encoder(name, **params):
    encoder = copy.deepcopy(encoders[name])
    encoder.set_params(**params)
    return encoder

if __name__ == '__main__':
    SinEncoder = Encoder(scale, sin, fixed_vector)
    SinEncoder.set_params(vmin=0.5, vmax=1.0, timesteps=100)

    print('SinEncoder params:', SinEncoder.get_params())

    SinGridEncoder = Encoder(grid, scale, sin, fixed_vector)
    grid = [[0,0,0],[0,1,0],[0,0,0]]
    SinGridEncoder.set_params(H0=0.5, H=1.0, grid=grid, timesteps=100, phi=10)

    print('SinGridEncoder params:', SinGridEncoder.get_params())

    input = np.array([[1,0,0]]).T
    print('input=', input)

    # SinEncoder.verbose = True
    # signal = SinEncoder(input)
    # print('signal=', signal)

    SinGridEncoder.verbose = True
    gsignal = SinGridEncoder(input)

    import matplotlib.pyplot as plt
    # plt.plot(signal)
    plt.subplot(211)
    for i in range(gsignal.shape[1]):
        for j in range(gsignal.shape[2]):
            plt.plot(gsignal[:,i,j,0], label=f'{i},{j}')
    plt.subplot(212)
    for i in range(gsignal.shape[1]):
        for j in range(gsignal.shape[2]):
            plt.plot(gsignal[:,i,j,1], label=f'{i},{j}')
    plt.legend()
    plt.show()



"""
Utilities related to the flatspin model
"""
import numpy as np
import copy
import itertools
import os
import inspect
from datetime import timedelta
from collections import OrderedDict

def numeric_dict(d):
    return {k: int(v) for k,v in d.items() if str(v).isnumeric()}

def filter_dict(d, keys):
    return d.__class__((k,v) for k,v in d.items() if k in keys)

def pop_dict(d, keys):
    return d.__class__((k, d.pop(k)) for k in list(d.keys()) if k in keys)

def try_eval(s):
    """ Try to eval(s), on failure return s """
    try:
        return eval(s, {'array': np.array})
    except:
        return s

def eval_dict(d, ctx=None):
    return d.__class__((k, try_eval(v)) for k,v in d.items())

def make_tuple(tup):
    if isinstance(tup, tuple):
        return tup
    return (tup,)

def unique_list(l):
    u = []
    for i in l:
        if not i in u:
            u.append(i)
    return u

def get_default_args(func, exclude=['self']):
    sig = inspect.signature(func)
    params = OrderedDict((k,p.default) for k,p in sig.parameters.items() if k not in exclude)
    return params

def get_default_kwargs(func, exclude=[]):
    params = get_default_args(func, exclude)
    params = OrderedDict((k,d) for k,d in params.items() if d is not inspect._empty)
    return params

def get_default_params(cls):
    try:
        default_params = cls.get_default_params()
    except AttributeError:
        if inspect.isclass(cls):
            # Walk up the class hierarchy to discover all available parameters
            # Assumes each sub class' __init__() will only contain new parameters
            # or modified defaults.  Old parameters are delegatet to the parent
            # class through **kwargs
            default_params = {}
            mro = inspect.getmro(cls)
            for cls in reversed(mro):
                default_params.update(get_default_kwargs(cls.__init__))
        else:
            default_params = get_default_kwargs(cls)
    return dict(default_params)

def check_params(params, classes):
    cls_params = split_params(params, classes)
    unknown_params = cls_params[-1]
    if unknown_params:
        raise TypeError("Invalid params", list(unknown_params.keys()))

def filter_params(cls, params):
    valid = get_default_params(cls)
    return filter_dict(params, valid)

def pop_params(cls, params):
    valid = get_default_params(cls)
    return pop_dict(params, valid)

def split_params(params, classes):
    """ Split params belonging to the given list of classes.

    Returns a list of parameters matching the the given list of classes. The
    last element of the list are parameters which didn't match any of the
    class signatures. """

    params = copy.deepcopy(params)
    valid_params = list(map(get_default_params, classes))

    # Check that classes don't have overlapping parameters
    overlap = set.intersection(*[set(p.keys()) for p in valid_params])
    assert not overlap, f"Overlapping parameters for {classes}: {overlap}"

    result = []
    for cls, valid in zip(classes, valid_params):
        cls_params = filter_dict(params, valid)
        result.append(cls_params)
        for k in cls_params.keys():
            del params[k]

    result.append(params) # remaining params

    return result

def format_label(label):
    if isinstance(label, np.ndarray):
        label = tuple(label)
    return str(label).replace(' ', '')

def label_columns(labels, prefix='spin'):
    labels = list(map(format_label, labels))
    return [prefix + str(c) for c in labels]

def label_columns_2d(labels, prefix='spin'):
    cols = label_columns(labels, prefix)
    cols = [[col + 'x', col + 'y'] for col in cols]
    return list(itertools.chain(*cols))

def get_outdir(script_filename):
    root, ext = os.path.splitext(script_filename)
    return root + ".out"

def make_outdir(script_filename):
    outdir = get_outdir(script_filename)
    os.makedirs(outdir, exist_ok=True)
    return outdir



"""
Dipole model runner
"""
import os
import numpy as np
import pandas as pd
import time
from numbers import Number
from datetime import timedelta
import subprocess
from copy import copy
from importlib import import_module
import warnings

from .cmdline import ProgressBar
from .encoder import get_encoder
from .data import read_table, to_table, save_table, is_archive, archive_key
from .utils import pop_params, label_columns, label_columns_2d, eval_dict

#
# List of deprecated parameters
#
# If you rename a parameter, please add it here to raise an error if it's being
# used! Otherwise it will go unnoticed since we ignore unknown parameters to
# facilitate fancy sweeps such as -s 'sz=arange(10,51,10)' -s 'size=[(sz,sz)]'
#
deprecated_params = ('sw_a', 'sw_alpha', 'temp_interp')

class DeprecationError(Exception):
    pass

def run(model, encoder='sin', input=1, input_key=None, periods=100, spp=100,
        temp=0, temp_func='interp', **params):
    """
    Run a flatspin simulation

    Parameters
    ----------
    model : SpinIce
        Model instance
    encoder : string or Encoder
        Name of input encoder to use.
        See flatspin.encoder for a list of available encoders
    input : float, array or string
        Input to be encoded
        float: constant input repeated periods times
        array: array of input to encode
        string: filename of table input data
    input_key : int or str
        Index into input (table column)
    periods : int
        Number of periods of the external field to run
        Only used if input is a float
    spp : int
        Number of samples to save per input value
    temp : float, array
        Temperature to be used
        float: constant temperature for whole run
        array: array of temperatures stretched to fit run
    params : dict
        Params to pass to the encoder
    """
    # unknown params are ignored

    if isinstance(encoder, str):
        encoder = get_encoder(encoder)
    encoder_params = pop_params(encoder, params)
    encoder.set_params(**encoder_params)

    # Bail out if user tries any deprecated params
    for p in params.keys():
        if p in deprecated_params:
            raise DeprecationError("Parameter is deprecated", p)

    # Warn on unknown params
    if params:
        unknown_params = ", ".join(params.keys())
        warnings.warn("Ignoring unknown parameters: " + unknown_params)

    result = {}

    if isinstance(input, str):
        # load input from file
        input = read_table(input)
        result['input'] = input
    elif isinstance(input, Number):
        # constant
        input = input * np.ones(periods)
        result['input'] = pd.DataFrame(input, columns=['input'])
    else:
        # array/list
        input = input
        result['input'] = pd.DataFrame(input, columns=['input'])

    if input_key is not None:
        input = input[input_key]
    input = np.array(input, dtype=float)

    if isinstance(encoder, str):
        encoder = get_encoder(encoder)
    encoder.set_params(**encoder_params)

    # np.random.seed(seed)
    # random.seed(seed)

    t0 = time.time()

    # Encode input as h_ext
    h_ext = encoder(input)

    # Figure out timesteps per period
    ts = len(h_ext) / len(input)
    if ts % spp != 0:
        print(f"WARNING: spp {spp} does not evenly divide the encoded input")
        print(f"         spp should divide the number of timesteps per input {ts}, e.g. spp={ts}")
    sample_every = int(np.round(ts / spp))

    # Input has now been encoded into h_ext, resulting in spp samples per input
    # value. Repeat input values spp times so that h_ext and input DataFrames
    # can be easily joined on the same index when needed. We keep the original
    # input index as an extra column called "index".
    input_df = result['input']
    input_df = input_df.loc[input_df.index.repeat(spp)].reset_index()
    input_df.index.name = 't'
    result['input'] = input_df

    # Thermal input
    temp = np.atleast_1d(temp)
    # exp-decay decays from temp[0] to temp[1] by nth order exponential decay 
    # where n is set by temp[2]
    if temp_func == 'exp-decay':
        exp_params = 3
        if len(temp) != exp_params:
            raise ValueError(f"For temp_func=exp, temp expects {exp_params}\
                parameters: temp=[start, stop, exponent]")
        lspace = np.linspace(0, 1, len(h_ext))
        temp = temp[0]*np.exp(np.log(temp[1]/temp[0])*pow(lspace, temp[2]))
    elif temp_func == 'interp':
        lspace = np.linspace(0, len(temp)-1, len(h_ext))
        temp = np.interp(lspace, np.arange(len(temp)), temp)
    else:
        if len(h_ext) % len(temp) != 0:
            raise ValueError(f"Length of temp {len(temp)} must divide length \
                of encoded input {len(h_ext)}")
        repeats = len(h_ext) / len(temp)
        temp = temp.repeat(repeats)

    save_h_therm = not np.all(temp == 0)

    result['geometry'] = pd.DataFrame({
        'posx': model.pos[:,0].copy(),
        'posy': model.pos[:,1].copy(),
        'angle': model.angle.copy()})

    result['hc'] = pd.DataFrame({'threshold': model.threshold.copy()})

    cols = label_columns(model.labels, prefix='init')
    result['init'] = pd.DataFrame([model.spin.copy()], columns=cols)

    # TODO: dump model?
    # make a copy of model so we store the initial state in the dataset
    #    'model': copy.deepcopy(model),

    steps = 0

    sample_index = []
    result['spin'] = []
    result['mag'] = []
    result['h_ext'] = []
    if save_h_therm:
        result['h_therm'] = []
    result['temp'] = []
    result['energy'] = []
    result['steps'] = []

    msg = type(model).__name__
    progress_bar = ProgressBar(desc=msg, total=len(h_ext))

    for i, (h, temp_std) in enumerate(zip(h_ext, temp)):
        if h.ndim > 1:
            # grid
            model.set_grid('h_ext', h)
        else:
            # global field
            model.set_h_ext(h)

        model.set_temperature(temp_std)

        steps += model.relax()

        progress_bar.update()

        if i % sample_every == 0:
            sample_index.append(i)
            result['spin'].append(model.spin.copy())
            result['mag'].append(model.vectors.flatten())
            result['energy'].append(model.total_energies())
            if save_h_therm:
                result['h_therm'].append(model.h_therm.flatten())
            result['temp'].append(model.thermal_std)
            if h.ndim > 1:
                # local field
                result['h_ext'].append(model.h_ext.flatten())
            else:
                # global field
                result['h_ext'].append(h)

            result['steps'].append(steps)

    progress_bar.close()

    cols = label_columns(model.labels, prefix='spin')
    result['spin'] = pd.DataFrame(result['spin'], columns=cols, index=sample_index, dtype=result['spin'][0].dtype)
    result['spin'].index.name = 't'

    cols = label_columns_2d(model.labels, prefix='mag')
    result['mag'] = pd.DataFrame(result['mag'], columns=cols, index=sample_index, dtype=result['mag'][0].dtype)
    result['mag'].index.name = 't'

    cols = ['h_extx', 'h_exty']
    if h_ext.ndim > 2:
        # local field
        cols = label_columns_2d(model.labels, prefix='h_ext')
    result['h_ext'] = pd.DataFrame(result['h_ext'], columns=cols, index=sample_index, dtype=model.h_ext.dtype)
    result['h_ext'].index.name = 't'

    if save_h_therm:
        cols = label_columns_2d(model.labels, prefix='h_therm')
        result['h_therm'] = pd.DataFrame(result['h_therm'], columns=cols, index=sample_index, dtype=np.float64)
        result['h_therm'].index.name = 't'

    result['temp'] = pd.DataFrame(result['temp'], columns=['temp'], index=sample_index)
    result['temp'].index.name = 't'

    cols = ["E_dip", "E_ext", "E_therm", "E_tot"]
    result['energy'] = pd.DataFrame(result['energy'], columns=cols, index=sample_index, dtype=np.float64)
    result['energy'].index.name = 't'

    result['steps'] = pd.DataFrame(result['steps'], columns=['steps'], index=sample_index, dtype=int)
    result['steps'].index.name = 't'


    runtime = time.time() - t0

    result['stats'] = {
        'runtime': runtime,
        'steps': steps,
    }

    td = timedelta(seconds=runtime)
    print(f'Completed {steps} steps in {td}')

    return result

def run_and_save(model_class, params, outdir, data_format):
    params_table = to_table(params)

    model_params = pop_params(model_class, params)
    model = model_class(**model_params)

    results = run(model, **params)
    results["params"] = params_table

    if not is_archive(outdir):
        os.makedirs(outdir)

    for name, data in results.items():
        df = to_table(data)
        if not is_archive(outdir):
            name = f"{name}.{data_format}"
        save_table(df, os.path.join(outdir, name))

def run_local(dataset):
    """ Run on localhost """
    n_runs = len(dataset)

    mod, cls = dataset.info['model'].rsplit('.', 1)
    module = import_module(mod)
    model_class = getattr(module, cls)
    data_format = dataset.info['data_format']

    for i, ds in enumerate(dataset):
        params = copy(ds.params)
        params.update(eval_dict(ds.row().to_dict()))
        outdir = params.pop('outdir')
        outpath = os.path.join(dataset.basepath, outdir)

        print(f"Run {i+1}/{n_runs}: {outdir}")
        print(params)
        run_and_save(model_class, params, outpath, data_format)

def generate_script(template, outfile, **params):
    with open(template) as fp:
        tpl = fp.read()
    script = tpl.format(**params)
    with open(outfile, 'w') as fp:
        fp.write(script)

job_script_template = os.path.join(os.path.dirname(__file__), 'flatspin.slurm.sh')

def run_dist(dataset, wait=True, max_jobs=1000):
    """ Run distributed on a cluster """

    #
    # Generate job script
    #

    # Construct a sensible name for the job script
    job_script_dir = dataset.basepath
    job_script_name = os.path.basename(job_script_template)
    job_script = os.path.join(job_script_dir, job_script_name)

    # Job template params
    job_params = {
        'job_script_dir': job_script_dir,
        'job_script_name': job_script_name,
        'basepath': dataset.basepath,
    }

    generate_script(job_script_template, job_script, **job_params)

    #
    # Submit jobs
    #

    # array size will never exceed max_jobs
    array_size = min(max_jobs, len(dataset))

    cmd = ['sbatch']
    if wait:
        cmd.append('--wait')
    cmd.append(f'--array=0-{array_size-1}')
    cmd.append(job_script)
    print(cmd)
    p = subprocess.Popen(cmd)

    if wait:
        p.wait()




import numpy as np
from .utils import approx

from flatspin import SpinIce
from numpy.testing import assert_array_equal

def test_init_no_temp(opencl):
    si = SpinIce(thermal_std=0, **opencl)
    assert si.h_therm is 0
    assert_array_equal(si.thermal_fields(),np.zeros((si.spin_count,2)))
    assert np.all([(si.thermal_field(i) is 0) for i in si.indices()])

def test_init_temp(opencl):
    si = SpinIce(thermal_std=1, **opencl)
    assert isinstance(si.h_therm, np.ndarray)
    assert np.all([isinstance(si.thermal_field(i),np.ndarray) for i in si.indices()])

    assert not si.thermal_fields() == approx(0)

def test_flip(opencl):
    si = SpinIce(thermal_std=1, **opencl)
    tf1 = si.thermal_fields()

    for i in range(0, si.spin_count, 2):
        si.flip(i)
    tf2 = si.thermal_fields()

    #field same for unchanged magnets
    assert tf1[range(1, si.spin_count, 2)] == approx(tf2[range(1, si.spin_count, 2)])
    #field flipped for flipped magnets
    assert tf1[range(0, si.spin_count, 2)] == approx(-tf2[range(0, si.spin_count, 2)])

def test_relax(opencl):
    si = SpinIce(thermal_std=1, **opencl)
    tf1 = si.thermal_fields()
    si.relax()

    assert si.thermal_fields() != approx(tf1)




"""
Vertex analysis
"""
import numpy as np
from skimage.util import view_as_windows

# List of patterns used to detect vertices
# Each element are spin angles on a grid
vertex_patterns = [
    np.deg2rad([[90], [90]]), # ising
    np.deg2rad([[0, 90, 0],[0, 0, 0], [0, 90, 0]]), # square closed
    np.deg2rad([[45.0, 135.0], [135.0, 45.0]]), # square open
    np.deg2rad([[0, -45, 0],[45, 0, 45], [0, -45, 0]]), # pinwheel diamond
    np.deg2rad([[0, 45, 0],[-45, 0, -45], [0, 45, 0]]), # pinwheel diamond
    np.deg2rad([[0, 45, 0],[135, 0, 135], [0, 45, 0]]), # pinwheel diamond
    np.deg2rad([[0, 135, 0],[45, 0, 45], [0, 135, 0]]), # pinwheel diamond
    np.deg2rad([[0, 90.0, 0], [-30.0, 0, 30.0]]), # kagome Y
    np.deg2rad([[30.0, 0, -30.0], [0, 90.0, 0]]), # kagome inverted Y
]

def find_vertices(grid, angle, win_size):
    """
    Find the vertices of a geometry

    Parameters
    ----------
    grid : Grid object
        The grid of the spin positions
    angle : 1D array
        The angles of each spin
    win_size : (height, width)
        The window size to scan the grid

    Returns a tuple (vi, vj, indices) where vi, vj are the vertex indices
    and indices is a list of spin indices corresponding to each vertex
    index.
    """
    index_grid = grid.map_values(np.arange(len(angle)), -1)
    angle_grid = grid.map_values(angle)
    angle_window = view_as_windows(angle_grid, win_size)
    index_window = view_as_windows(index_grid, win_size)

    is_vertex = np.zeros(angle_window.shape[:2], dtype=bool)
    for pattern in vertex_patterns:
        if pattern.shape == win_size:
            # compatible pattern
            match = np.all(angle_window == pattern, axis=(2,3))
            is_vertex = np.logical_or(is_vertex, match)

    vertex_indices = is_vertex.nonzero()

    spin_indices = index_window[is_vertex]
    spin_indices = [i[i>=0] for i in spin_indices]

    return vertex_indices + (spin_indices,)

def vertex_type_ising(spin, angle):
    if spin[0] == spin[1]:
        return 1
    return 2

def vertex_type_square(spin, angle):
    dir_in = [1, 1, -1, -1]
    spin = spin * dir_in
    if spin[0] == -1:
        # flipping the spins in a vertex won't change its type
        spin *= -1
    num_in = np.count_nonzero(spin == 1)

    if num_in == 2:
        if tuple(spin) == (1, -1, -1, 1):
            # Type I
            return 1
        # Type II
        return 2

    if num_in == 3 or num_in == 1:
        # Type III
        return 3

    # Type IV
    return 4

def vertex_type_tri(spin, angle):
    if angle[-1] == np.pi/2:
        # inverted Y vertex
        dir_in = [1, -1, -1]
    else:
        # Y vertex
        dir_in = [1, 1, -1]

    spin *= dir_in
    if spin[0] == -1:
        # flipping the spins in a vertex won't change its type
        spin *= -1

    num_in = np.count_nonzero(spin == 1)
    if num_in == 3:
        # Type II
        return 2

    # Type I
    return 1

def vertex_type(spin, angle):
    """ Determine the type of a vertex given its spins and angles """
    if len(spin) == 2:
        return vertex_type_ising(spin, angle)
    elif len(spin) == 3:
        return vertex_type_tri(spin, angle)
    elif len(spin) == 4:
        return vertex_type_square(spin, angle)

    raise NotImplementedError(f"Don't know about vertices of size {len(spin)}, sorry.")

def vertex_pos(pos, vertices):
    return np.array([pos[v].mean(axis=0) for v in vertices])

def vertex_mag(mag, vertices):
    return np.array([mag[v].sum(axis=0) for v in vertices])




#!/usr/bin/env python3
import sys
import os
import argparse
import shlex
import pandas as pd
from collections import OrderedDict
from copy import copy

from flatspin.cmdline import StoreKeyValue, eval_params
from flatspin.sweep import sweep
from flatspin.data import Dataset, table_formats, is_archive_format
from flatspin.runner import run, run_local, run_dist
from flatspin.encoder import encoders
from flatspin.utils import get_default_params
from .run import import_model, main_list_params

def main():
    parser = argparse.ArgumentParser(description='Run flatspin sweep')
    parser.add_argument('-r', '--run', choices=['local', 'dist', 'none'], default='local',
                        help='run locally or distributed on a cluster')
    parser.add_argument('-p', '--param', action=StoreKeyValue, default={},
                        help='set model parameter key=value')
    parser.add_argument('-s', '--sweep', action=StoreKeyValue, metavar='key=SPEC',
                        help='set sweep parameter key=SPEC')
    parser.add_argument('-n', '--repeat', type=int, default=1, metavar='N',
                        help='repeat each experiment N times (default: %(default)s)')
    parser.add_argument('-ns', '--repeat-spec', action=StoreKeyValue, metavar='key=SPEC',
                        help='repeat each according to key=SPEC')
    parser.add_argument('--seed', type=int, default=0, help='random seed')
    parser.add_argument('-m', '--model', default='SquareSpinIceClosed',
            help='name of model (default: %(default)s) ')
    parser.add_argument('-e', '--encoder', choices=encoders.keys(), default='sin',
            help='input encoder (type of external field, default: %(default)s)')
    parser.add_argument('-o', '--basepath', help='output directory for results')
    parser.add_argument('-f', '--format', choices=table_formats.keys(), default='npz',
            help='format of output files (default: %(default)s)')
    parser.add_argument('--list-params', action='store_true')

    args = parser.parse_args()

    if args.list_params:
        return main_list_params(args)

    if args.basepath is None or args.sweep is None:
        parser.error('The following arguments are required: -o/--basepath, -s/--sweep')

    model_class = import_model(args.model)
    model_name = model_class.__name__
    encoder_name = args.encoder
    encoder = encoders[encoder_name]

    # -e sin is really just an alias for -p encoder=sin
    params = get_default_params(run)
    params['encoder'] = encoder_name
    params.update(get_default_params(model_class))
    params.update(get_default_params(encoder))
    params.update(eval_params(args.param))

    info = {
        'model': f'{model_class.__module__}.{model_class.__name__}',
        'model_name': model_name,
        'data_format': args.format,
        'command': ' '.join(map(shlex.quote, sys.argv)),
    }

    ext = args.format if is_archive_format(args.format) else "out"
    if args.repeat > 1 or args.repeat_spec:
        outdir_tpl = model_name + ".{:06d}.{:06d}." + ext
    else:
        outdir_tpl = model_name + ".{:06d}." + ext

    basepath = args.basepath
    if os.path.exists(basepath):
        # Refuse to overwrite an existing dataset
        raise FileExistsError(basepath)
    os.makedirs(basepath)

    index = []

    # Generate index
    for i, j, sweep_params in sweep(args.sweep, args.repeat, args.repeat_spec, params, args.seed):

        outdir = outdir_tpl.format(i, j)
        outpath = os.path.join(basepath, outdir)

        row = OrderedDict(sweep_params)
        row.update({'outdir': outdir})
        index.append(row)

    # Save dataset
    index = pd.DataFrame(index)
    dataset = Dataset(index, params, info, basepath)
    dataset.save()

    # Run!
    print("Starting sweep with {} runs".format(len(dataset)))
    if args.run == 'local':
        run_local(dataset)

    elif args.run == 'dist':
        run_dist(dataset)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
import os
import numpy as np
import pandas as pd
from numpy.linalg import norm
from math import ceil, sqrt
from scipy.spatial import cKDTree
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize, LinearSegmentedColormap
from matplotlib.collections import PatchCollection
import random
import time
import itertools
from datetime import timedelta
from copy import copy
from PIL import Image
import pkg_resources

from flatspin.grid import Grid
from flatspin.vertices import find_vertices, vertex_type
from flatspin.utils import get_default_params
from flatspin.plotting import plot_vectors

normalize_rad = Normalize(vmin=0, vmax=2*np.pi)
normalize_vertex_type = Normalize(vmin=1, vmax=4)
# Type 1: green, type 2: blue, type 3: red, type 4: gray
vertex_clist = [(0,1,0), (0,0,1), (1,0,0), (.5,.5,.5)]
vertex_cmap = LinearSegmentedColormap.from_list("vertex", vertex_clist, N=4)

class LabelIndexer:
    def __init__(self, labels):
        labels = np.array(labels)
        if len(labels.shape) == 1:
            labels.shape += (1,)

        self.index = pd.MultiIndex.from_arrays(labels.T)

    def __getitem__(self, label):
        try:
            index = self.index.get_locs(label)
        except TypeError:
            # single label
            index = self.index.get_locs([label])

        if len(index) == 1:
            # single index
            return index[0]

        # list of indices
        return index

    def __repr__(self):
        return repr(self.index)

class SpinIce(object):
    """ Ising spin lattice """
    def __init__(self, *, size=(4,4), lattice_spacing=1, hc=10, alpha=1.0,
            disorder=0, h_ext=(0,0), neighbor_distance=1, switching='sw',
            sw_b=1, sw_c=1, sw_beta=3, sw_gamma=3, thermal_std=0,
            flip_mode='max', init='polarized',
            random_prob=0.5, random_seed=None,
            use_opencl=False, opencl_platform=0, opencl_device=0):
        """
        Parameters
        ----------
        size : tuple of int
            The size of the spin ice (width, height).
            Note that the size is geometry specific.
        lattice_spacing : float
            Spacing between each spin.
        hc : float
            Mean switching threshold.
        disorder : float
            Switching threshold disorder as percentage of hc. If non-zero,
            sample switching from a normal distribution with mean hc and
            standard deviation hc * disorder.
        h_ext : tuple of float
            External field (h_extx, h_exty).
        neighbor_distance : float
            Neighborhood to consider when calculating dipole interactions.
            All spins within lattice_spacing * neighbor_distance are considered
            neighbors. Set neighbor_distance=np.inf for a global neighborhood.
        switching : {'sw', 'budrikis'}
            Magnetic switching model:
                - 'sw': Extended Stoner-Wohlfarth model
                - 'budrikis': switching based on the paralell field component
        sw_b : float
            Height of the Stoner-Wohlfarth astroid (paralell axis).
        sw_c : float
            Width of the Stoner-Wohlfarth astroid (perpendicular axis).
        sw_beta : float
            Pointiness of the top/bottom of the Stoner-Wohlfarth astroid
            (paralell axis).
        sw_gamma : float
            Pointiness of the left/right of the Stoner-Wohlfarth astroid
            (perpendicular axis).
        thermal_std : float (positive)
            Standard deviation of the gaussian distribution which the thermal noise is sampled 
        flip_mode : {'single', 'max', 'max-rand', 'all'}
            Flip mode strategy:
                - 'single': single uniform random flip (Budrikis Monte Carlo)
                - 'max': flip the spin with maximum energy, break ties by index
                - 'max-rand': flip the spin with maximum energy, break ties randomly
                - 'all': flip all flippable spins
        init : {'polarized', 'random', 'image.png'} or ndarray
            Initial spin state.
        use_opencl : bool
            Use OpenCL acceleration.
        opencl_platform : int
            Select OpenCL platform (if multiple GPUs of different families are installed).
        opencl_device : int
            Select OpenCL device (if multiple GPUs of the same family are installed).
        """
        self.size = size
        self.lattice_spacing = lattice_spacing
        self.hc = hc
        self.alpha = alpha
        self.disorder = disorder
        self.switching = switching
        self.sw_params = (sw_b, sw_c, sw_beta, sw_gamma)
        self.thermal_std = thermal_std
        self.flip_mode = flip_mode
        self.neighbor_distance = neighbor_distance
        self.random_prob = random_prob
        assert switching in ('budrikis', 'sw')
        assert flip_mode in ('single', 'max', 'max-rand', 'all')
        assert neighbor_distance > 0

        self.labels = None

        self.cl = use_opencl
        self.cl_device = opencl_device
        self.cl_platform = opencl_platform
        self._cl_context = None

        if random_seed is not None:
            np.random.seed(random_seed)

        # geometry
        pos, angle = self._init_geometry()
        self.pos = pos.astype(float)
        self.angle = angle.astype(float)
        self.m = np.column_stack([np.cos(self.angle), np.sin(self.angle)])
        self.spin_count = len(self.pos)

        if self.labels is None:
            # No labels, labels = indices
            self.labels = np.arange(self.spin_count)

        self.L = LabelIndexer(self.labels)

        # thermal noise
        self.update_thermal_noise()

        # spin flip threshold
        self.threshold = np.zeros(self.spin_count)
        self._init_threshold()

        # spin state
        self.spin = np.ones(self.spin_count, dtype=np.int8)
        self._init_spin(init)

        # flat neighbor list (computed once)
        self._neighbor_list = self._init_neighbor_list()
        self.num_neighbors = self._neighbor_list.shape[-1]

        # dipolar fields between each spin and its neighbors
        # precomputed on demand by _h_dip/_init_cl
        self._h_dip_cache = None

        self.set_h_ext(h_ext)

    @property
    def N(self):
        """ Alias for self.spin_count """
        return self.spin_count

    def label(self, i):
        """ Get the label of a given spin index or list of indices """
        return tuple(self.labels[i])

    def indexof(self, label):
        """ Get the spin index of a given label or range of labels

        Alias for self.L[label]
        """
        return self.L[label]

    def indices(self):
        """ Get all spin indices """
        return range(self.spin_count)

    def all_indices(self):
        """ Get all spin indices as numpy integer array index
        See https://docs.scipy.org/doc/numpy/reference/arrays.indexing.html#integer-array-indexing
        """
        return list(self.indices())

    def _init_spin(self, init):
        if isinstance(init, str) and init == 'random':
            self.randomize(self.random_prob)

        if isinstance(init, str) and init.endswith('png'):
            self.set_spin_image(init)

        if isinstance(init, np.ndarray):
            self.set_spin(init)

    def _init_geometry(self):
        nx, ny = self.size
        spin_count = nx * ny
        pos = np.zeros((spin_count, 2), dtype=float)
        angle = np.zeros(spin_count, dtype=float)

        # rotation of each spin
        angle[:] = np.pi/2

        # positions of each spin
        rows = np.arange(ny, dtype=int)
        cols = np.arange(nx, dtype=int)
        grid = np.array(np.meshgrid(rows, cols))
        grid = grid.T.reshape((-1, 2))
        pos = self.lattice_spacing * np.flip(grid.astype(float), axis=-1)
        pos[:] = pos
        self.labels = grid

        return pos, angle
        
    def _init_threshold(self):
        # spin flip threshold
        if self.disorder:
            std = self.hc * self.disorder
            threshold = np.random.normal(self.hc, std, self.threshold.shape)
            # all thresholds should be positive
            threshold = np.abs(threshold)
            self.threshold[:] = threshold
        else:
            self.threshold[:] = self.hc

    def _init_neighbor_list(self):
        # Calculate neighborhood matrix
        # print(f'_init_neighbor_list(size={self.size})')
        neighbors = []
        num_neighbors = 0

        # Construct KDTree for every position
        tree = cKDTree(self.pos)

        nd = self.lattice_spacing * self.neighbor_distance
        nd += 1e-5 # pad to avoid rounding errors

        for i in self.indices():
            p = self.pos[i]
            n = tree.query_ball_point([p], nd)[0]
            n.remove(i)
            neighbors.append(n)
            num_neighbors = max(num_neighbors, len(n))

        # print(f'num_neighbors={num_neighbors}')

        # Neighborhood list, -1 marks end of each list
        neighbor_list = np.full((self.spin_count, num_neighbors), -1, dtype=np.int32)
        for i, neighs in enumerate(neighbors):
            neighbor_list[i,:len(neighs)] = neighs

        return neighbor_list

    def _init_h_dip(self):
        # when use_opencl=1, the hdip cache is initialized by _init_cl()
        assert not self.cl

        # cache dipolar fields for the neighborhood of all magnets
        _h_dip = np.zeros((self.spin_count, self.num_neighbors, 2), dtype=np.float64)
        for i in self.indices():
            for jj, j in enumerate(self.neighbors(i)):
                # print(f'i={i} jj={jj} j={j}')
                _h_dip[i][jj] = self.alpha * np.array(self.spin_dipolar_field(i, j))

        return _h_dip

    def __getstate__(self):
        odict = copy(self.__dict__)
        # Don't pickle _cl attributes. They will be automatically
        # re-initialized by _init_cl() when needed
        for k in list(filter(lambda k: k.startswith('_cl'), odict)):
            odict[k] = None
        return odict

    def __eq__(self, other):
        state = self.__getstate__()
        other_state = other.__getstate__()
        for k in state:
            if np.any(state[k] != other_state[k]):
                return False
        return True

    @property
    def width(self):
        x = self.pos[:, 0]
        xmin, xmax = np.min(x), np.max(x)
        return xmax - xmin

    @property
    def height(self):
        y = self.pos[:, 1]
        ymin, ymax = np.min(y), np.max(y)
        return ymax - ymin

    @property
    def vectors(self):
        """ Spin vectors """
        spin = np.array(self.spin, dtype=float)
        spin.shape += (1,)
        mag = self.m * spin
        return mag

    def set_spin(self, spin):
        spin = np.array(spin, dtype=np.int8)
        assert spin.shape == (self.spin_count,)
        assert np.all(np.logical_or(spin == 1, spin == -1))
        self.spin = spin

    def set_spin_image(self, filename):
        """ Set spin state from image

        The image is resized to fit the default spin grid. Then the grayscale
        image is used to set spin state:
        * black maps to spin -1
        * non-black maps to spin 1
        """
        # Use default grid
        G = self.grid()
        # calculate size of grid
        height, width = G.size

        # load image
        img = Image.open(filename)
        img = img.convert('L') # convert to grayscale

        # flip it since (0,0) is top left in the image but (0,0) is bottom left
        # in our spin geometry
        img = img.transpose(Image.FLIP_TOP_BOTTOM)

        # resize image to the default grid size
        img = img.resize((width, height))
        img = np.array(img)

        img[img != 0] = 1  # map non-black to spin 1
        img[img == 0] = -1 # map black to spin -1

        # finally set all the spins
        self.set_grid('spin', np.array(img))

    def set_threshold(self, threshold):
        assert threshold.shape == (self.spin_count,)
        self.threshold = threshold

    def set_h_ext(self, h_ext):
        """
        Set external field to h_ext

        h_ext can either be a single vector for a uniform field (h_ext.shape==(2,))
        or a 2D array of vectors for a non-uniform field (h_ext.shape==(self.spin_count, 2))
        """
        h_ext = np.array(h_ext, dtype=np.float64)
        assert h_ext.shape == (2,) or h_ext.shape == (self.spin_count, 2)
        if h_ext.shape == (2,):
            h_ext = np.tile(h_ext, (self.spin_count, 1))
        self.h_ext = h_ext

    def set_h_ext_grid(self, h_ext):
        if h_ext.shape == (2,):
            # global
            self.set_h_ext(h_ext)
        else:
            # grid
            self.set_grid('h_ext', h_ext)

    def set_temperature(self,std):
        self.thermal_std = std

    def randomize(self, prob=0.5):
        self.spin = -1+2*np.random.binomial(1, 1-prob, self.spin_count)
        self.spin = self.spin.astype(np.int8)
        #self.spin = -1+2*np.random.randint(0, 2, self.shape, dtype=np.int8)

    def polarize(self):
        self.spin[:] = 1

    def update_thermal_noise(self):
        """ Resamples thermal noise.
        Samples x and y components of noise from normal distribution"""
        if self.thermal_std:
            self.h_therm = np.random.normal(0, self.thermal_std, (self.spin_count, 2))
        else:
            self.h_therm = 0

    def neighbors(self, i):
        neighs = self._neighbor_list[i]
        return neighs[neighs >= 0]

    def spin_dipolar_field(self, i, j):
        """ Calculate dipolar field between spin i and j relative to positive spin """
        r = self.pos[j] - self.pos[i]
        m = self.m
        dist = norm(r)
        h_dip_1 = -m[j] / dist**3
        h_dip_2 = 3 * r * m[j].dot(r) / dist**5
        h_dip = h_dip_1 + h_dip_2

        m_angle = np.arctan2(m[i][1], m[i][0])
        h_angle = np.arctan2(h_dip[1], h_dip[0])
        h = norm(h_dip)
        theta = h_angle - m_angle
        h_par = h*np.cos(theta)
        h_perp = h*np.sin(theta)
        return np.array([h_par, h_perp], dtype=float)

    def dipolar_field(self, i):
        """ Calculate total dipolar field parallell to spin i """
        if self.cl:
            # path only for testing
            return self._h_dip_local_cl()[i]
        return self._h_dip_local(i)

    def dipolar_fields(self):
        if self.cl:
            return self._h_dip_local_cl()

        h_dip = np.zeros((self.spin_count, 2))
        for i in self.indices():
            h_dip[i] = self.dipolar_field(i)
        return h_dip

    @property
    def _h_dip(self):
        if self._h_dip_cache is None:
            self._h_dip_cache = self._init_h_dip()
        return self._h_dip_cache

    def _h_dip_local(self, i):
        hdip = np.zeros(2) # h_par, h_perp
        for jj, j in enumerate(self.neighbors(i)):
            # print("h_dip", jj, j)
            hdip += self._h_dip[i][jj] * self.spin[i] * self.spin[j]
        return hdip

    def external_field(self, i):
        """ Calculate external field parallel and perpendicular to spin i """
        if self.cl:
            # path only for testing
            return tuple(self._external_fields_cl()[i])

        m = self.spin[i] * self.m[i]
        m_angle = np.arctan2(m[1], m[0])

        h_ext = self.h_ext[i]
        h = norm(h_ext)
        h_angle = np.arctan2(h_ext[1], h_ext[0])

        theta = h_angle - m_angle
        h_par = h*np.cos(theta)
        h_perp = h*np.sin(theta)

        return np.array([h_par, h_perp], dtype=np.float64)

    def external_fields(self):
        """ Calculate external fields parallel and perpendicular to all spins """
        if self.cl:
            return self._external_fields_cl()

        # todo: numpyfy
        h_ext = np.zeros((self.spin_count, 2))
        for i in self.indices():
            h_ext[i] = self.external_field(i)
        return h_ext

    def thermal_field(self, i):
        if self.h_therm is 0:
            return 0
        return (self.h_therm[i] * self.spin[i])

    def thermal_fields(self):
        if self.h_therm is 0:
            return np.zeros((self.spin_count, 2))
        return (self.h_therm.T * self.spin).T

    def total_field(self, i):
        """ Calculate the total field parallel to spin i """
        if self.cl:
            return self._total_fields_cl()[i] + self.thermal_field(i)
        return self.dipolar_field(i) + self.external_field(i) + self.thermal_field(i)

    def total_fields(self):
        therm_f = 0 if self.h_therm is 0 else self.thermal_fields() #TODO: make thermal_fields_cl
        if self.cl:
            return self._total_fields_cl() + therm_f
        return self.dipolar_fields() + self.external_fields() +  therm_f

    def flip(self, i):
        self.spin[i] *= -1

    # TODO: rename / discard?
    def _switching_energy_budrikis(self):
        h_tot = self.total_fields()
        h_par = h_tot[..., 0] # paralell components only
        E = -(h_par + self.threshold)
        return E

    def _switching_energy_sw(self):
        b, c, beta, gamma = self.sw_params

        h_tot = self.total_fields()
        h_par = h_tot[..., 0] / (b * self.threshold)
        h_perp = h_tot[..., 1] / (c * self.threshold)

        # Switching criteria 1: h_par**(2/3) + h_perp**(2/3) > hc**(2/3)
        # E > 0: switching is possible
        E = (h_par**2)**(1/gamma) + (h_perp**2)**(1/beta) - 1

        # Switching criteria 2: h_par < 0
        # make E negative when h_par >= 0
        # Use a very small number here instead of zero to avoid a corner case:
        # when the field is perfectly perpendicular to the magnet, rounding
        # errors can cause h_par to always be negative regardless of the spin,
        # resulting in an infinite switching loop.
        cond = h_par >= -1e-12
        E[cond] = -np.abs(E[cond])
        return E

    def switching_energy(self):
        if self.switching == 'budrikis':
            E = self._switching_energy_budrikis()
        else:
            E = self._switching_energy_sw()

        return E

    def flippable_energy(self):
        flippable = []
        energy = self.switching_energy()
        flippable = np.nonzero(energy > 0)
        energy = energy[flippable]
        flippable = flippable[0]
        return flippable, energy

    def flippable(self):
        return self.flippable_energy()[0]

    def step(self):
        """ Perform a flip of one or more flippable spins """
        flip, energy = self.flippable_energy()
        # print("flippable:", len(flip))
        if len(flip) == 0:
            return False

        # TODO: rename / parameterize?
        if self.flip_mode == 'single':
            # Glip a single random spin
            i = np.random.choice(flip)
            self.flip(i)

        elif self.flip_mode == 'max':
            # Flip the spin with maximum energy
            i = flip[np.argmax(energy)]
            self.flip(i)

        elif self.flip_mode == 'max-rand':
            # Flip the spin with maximum energy, ties broken by random choice
            # Break ties by random choice
            i = np.argmax(energy)
            ties, = np.nonzero(energy == energy[i])
            j = flip[np.random.choice(ties)]
            self.flip(j)

        # TODO: drop?
        elif self.flip_mode == 'all':
            for i in flip:
                self.flip(i)

        return True

    def relax(self):
        """ Flip spins until equilibrium, return number of calls to step() """
        steps = 0

        self.update_thermal_noise()

        while self.step():
            steps += 1
        return steps

    def energy(self):
        h_tot = self.total_fields()
        h_par = h_tot[..., 0] # paralell components only
        E = -h_par
        return E

    def total_energies(self):
        dip = -np.sum(self.dipolar_fields()[...,0])
        ext = -np.sum(self.external_fields()[...,0])

        tf = self.thermal_fields()
        therm = -np.sum(tf[...,0]) if tf is not 0 else 0

        total = np.sum((dip,therm,ext))
        return [dip, ext, therm, total]

    def total_energy(self):
        return np.sum(self.energy())

    def total_magnetization(self):
        mag = self.vectors
        return np.array(np.sum(mag, axis=0))

    # sub-classes can override this to specify the size of the vertex window
    # (see find_vertices)
    _vertex_size = (2, 1)

    def find_vertices(self):
        """ Find the vertices in this geometry

        Returns a tuple (vi, vj, indices) where vi, vj are the vertex indices
        and indices is a list of spin indices corresponding to each vertex
        index.
        """
        return find_vertices(self.grid(), self.angle, self._vertex_size)

    def vertices(self):
        """ Get the spin indices of all vertices """
        vi, vj, indices = self.find_vertices()
        return indices

    def vertex_indices(self):
        """ Get all vertex indices """
        vi, vj, indices = self.find_vertices()
        return vi, vj

    def vertex_type(self, v):
        """ Get the vertex type for a vertex where v are the spin indices
        of the vertex """
        return vertex_type(self.spin[v], self.angle[v])

    def vertex_count(self):
        """ Count the number of vertices of each type

        Returns a tuple (types, counts) where types are the different vertex
        types and counts are the corresponding counts. """
        vertex_types = [self.vertex_type(v) for v in self.vertices()]
        types, counts = np.unique(vertex_types, return_counts=True)
        return tuple(types), tuple(counts)

    def vertex_population(self):
        """ Calculate the vertex type population as a fraction of all vertices

        Returns a tuple (types, pops) where types are the different vertex
        types and pops are the corresponding fractions. """
        types, counts = self.vertex_count()
        counts = np.array(counts)
        pops = counts / np.sum(counts)
        return types, tuple(pops)

    def vertex_pos(self, v):
        """ Get the position of a vertex where v are the spin indices of the
        vertex """
        return np.mean(self.pos[v], axis=0)

    def vertex_mag(self, v):
        """ Get the direction of a vertex v """
        spin = self.spin[v]
        spin.shape += (1,)
        m = self.m[v]
        mag = m * spin
        return np.sum(mag, axis=0)

    @property
    def _default_cell_size(self):
        return (self.lattice_spacing, self.lattice_spacing)

    def grid(self, cell_size=None):
        """ Map spin indices onto a regular grid

        The spacing between each grid point is given by cell_size.
        If cell_size <= lattice_spacing, each grid cell will contain at
        most one spin.
        If cell_size > lattice_spacing, each grid cell may contain more
        than one spin.
        If cell_size is None, an optimal (geometry dependent) cell size is used

        Returns a Grid object which allows quick lookup of spin index to
        grid index
        """
        if cell_size is None:
            cell_size = self._default_cell_size
        if np.isscalar(cell_size):
            cell_size = (cell_size, cell_size)

        # Calculate optimal padding
        padx = min(self._default_cell_size[0]/2, cell_size[0]/2)
        pady = min(self._default_cell_size[1]/2, cell_size[1]/2)
        padding = (padx, pady)

        grid = Grid(self.pos, cell_size, padding)

        return grid

    def fixed_grid(self, grid_size):
        """ Map spin indices onto a regular grid of fixed size

        Like grid() but takes grid size (number of cells) as parameter
        instead of cell size.
        """
        padx = self._default_cell_size[0]/2
        pady = self._default_cell_size[1]/2
        xmax = np.max(self.pos[..., 0])
        ymax = np.max(self.pos[..., 1])
        width = xmax + 2 * padx
        height = ymax + 2 * pady
        cell_size = (width / grid_size[0]), (height / grid_size[1])
        return self.grid(cell_size)

    def set_grid(self, attr, values):
        """ Map grid values onto some spin attribute.

        Valid attributes: spin, h_ext
        """
        G = self.fixed_grid((values.shape[1], values.shape[0]))

        # spin indices as numpy integer array index
        spin_inds = self.all_indices()

        # map all spin indices to grid index
        grid_inds = G.grid_index(spin_inds)

        # determine shape of attr and the corresponding set-function
        set_attr = getattr(self, 'set_' + attr)
        oldattr = getattr(self, attr)
        newattr = oldattr.copy()
        newattr[spin_inds] = values[grid_inds]
        set_attr(newattr)

    def view_grid(self, attr, cell_size=None, method='sum'):
        """ Project some spin attribute onto a grid.

        Valid attributes: spin, vectors, h_ext, threshold, pos

        The spacing between each grid cell is given by cell_size.
        If cell_size <= lattice_spacing, each grid cell will contain at
        most one spin attribute.
        If cell_size > lattice_spacing, each grid cell will contain the
        sum of several spin attributes.
        """

        # make the grid
        G = self.grid(cell_size)

        # get the values we wish to view on the grid
        values = getattr(self, attr)

        return G.add_values(values, method=method)

    # TODO: return spin instead of vectors
    def spin_grid(self, cell_size=None):
        """ Project the spin vectors onto a grid.

        See view_grid() for information about cell_size
        """
        return self.view_grid('vectors', cell_size)

    def plot(self, arrows=True, **kwargs):
        return plot_vectors(self.pos, self.vectors, arrows=arrows, **kwargs)

    def plot_energy(self, arrows=True):
        E = self.energy()
        return plot_vectors(self.pos, self.vectors, C=E, arrows=arrows, cmap='bwr')

    def plot_vertices(self):
        size = self.lattice_spacing * 0.15
        vertices = self.vertices()
        circles = [plt.Circle(self.vertex_pos(v), size) for v in vertices]
        colors = np.array([self.vertex_type(v) for v in vertices])
        col = PatchCollection(circles, cmap=vertex_cmap, norm=normalize_vertex_type)
        col.set_array(colors)
        ax = plt.gca()
        ax.add_collection(col)
        ax.autoscale_view()
        return col

    def plot_vertex_mag(self, arrows=True, **kwargs):
        vertices = self.vertices()
        XY = np.array([self.vertex_pos(v) for v in vertices])
        UV = np.array([self.vertex_mag(v) for v in vertices])

        # Normalize vectors to unit length
        nmax = norm(UV, axis=-1).max()
        if nmax != 0:
            UV = UV / nmax

        return plot_vectors(XY, UV, arrows=arrows, **kwargs)

    def _init_h_dip_cl(self):
        # Only called by _init_cl()
        posr = self.pos.ravel()
        pos_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=posr)
        res_g = cl.Buffer(self._cl_context, clmf.WRITE_ONLY, 2 * self.spin_count * self.num_neighbors * np.dtype(np.float64).itemsize)
        self._cl_prg.spin_dipolar_field(
                self._cl_queue,
                (self.spin_count, self.num_neighbors),
                None,
                pos_g, 
                res_g,
                self._cl_neighbors_g,
                self._cl_n_neighbors_g,
                self._cl_m_g
                )


        res = np.empty((self.spin_count, self.num_neighbors, 2), dtype=np.float64)

        cl.enqueue_copy(self._cl_queue, res, res_g)

        return self.alpha * res

    def _init_cl(self):
        assert self.cl
        if self._cl_context:
            # already initialized
            return

        global cl, clmf
        cl = __import__("pyopencl")
        clmf = cl.mem_flags

        platforms = cl.get_platforms()
        assert(len(platforms) > 0), "Couldn't find any platforms! Driver installed?"
        platform = platforms[self.cl_platform]

        devices = platform.get_devices()
        assert(len(devices) > 0), "No devices in platform"
        device = devices[self.cl_device]
        print("Using device {} on {}".format(device.name, platform.name))

        self._cl_context = cl.Context([device])
        self._cl_queue = cl.CommandQueue(self._cl_context)

        src = pkg_resources.resource_string(__name__, "flatspin_cl_kernels.c")
        src = src.decode("utf-8")

        self._cl_prg = cl.Program(self._cl_context, src).build(options=[])

        #preload some of the buffers.
        self._cl_spin_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=self.spin)
        self._cl_res_g = cl.Buffer(self._cl_context, clmf.WRITE_ONLY,
                np.prod((self.spin_count, 2)) * np.dtype(np.float64).itemsize)

        self._cl_neighbors_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=self._neighbor_list)
        self._cl_n_neighbors_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=np.int32(self.num_neighbors))

        self._cl_m_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=np.float64(self.m))
        self._cl_h_ext_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=self.h_ext)

        # we are now ready to calculate h_dip cache on the GPU
        if self._h_dip_cache is None:
            self._h_dip_cache = self._init_h_dip_cl()

        self._cl_hdip_cache_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=self._h_dip_cache)

    def _total_fields_cl(self):
        self._init_cl()

        #this will calcualte for all indices
        #copy the updated data to the device
        #print(self.spin.shape)

        cl.enqueue_copy(self._cl_queue, self._cl_spin_g, self.spin)
        cl.enqueue_copy(self._cl_queue, self._cl_h_ext_g, self.h_ext)

        self._cl_prg.total_fields(self._cl_queue, (self.spin_count,), None,
                self._cl_spin_g,
                self._cl_hdip_cache_g,
                self._cl_res_g,
                self._cl_neighbors_g,
                self._cl_n_neighbors_g,
                self._cl_m_g,
                self._cl_h_ext_g
        )

        res = np.empty((self.spin_count, 2), dtype=np.float64)
        cl.enqueue_copy(self._cl_queue, res, self._cl_res_g)

        return res

    def _h_dip_local_cl(self):
        self._init_cl()

        #this will calcualte for all indices
        cl.enqueue_copy(self._cl_queue, self._cl_spin_g, self.spin)


        self._cl_prg.test_h_dip_local(self._cl_queue, (self.spin_count,), None,
                self._cl_spin_g,
                self._cl_hdip_cache_g,
                self._cl_res_g,
                self._cl_neighbors_g,
                self._cl_n_neighbors_g
        )

        res = np.empty((self.spin_count, 2), dtype=np.float64)
        cl.enqueue_copy(self._cl_queue, res, self._cl_res_g)

        return res

    def _external_fields_cl(self):
        self._init_cl()

        #this will calcualte for all indices
        spin_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=self.spin)

        cl.enqueue_copy(self._cl_queue, self._cl_spin_g, self.spin)
        cl.enqueue_copy(self._cl_queue, self._cl_h_ext_g, self.h_ext)


        self._cl_prg.test_h_ext(self._cl_queue, (self.spin_count,), None,
                self._cl_spin_g,
                self._cl_m_g,
                self._cl_h_ext_g,
                self._cl_res_g,
        )

        res = np.empty((self.spin_count, 2), dtype=np.float64)
        cl.enqueue_copy(self._cl_queue, res, self._cl_res_g)

        return res



class SquareSpinIce(SpinIce):
    pass

class SquareSpinIceClosed(SquareSpinIce):
    def __init__(self, *, lattice_spacing=2*np.cos(np.pi/4), edge="symmetric", **kwargs):
        kwargs['lattice_spacing'] = lattice_spacing

        assert edge in ("symmetric", "asymmetric")
        self.edge = edge

        super().__init__(**kwargs)

    def _init_geometry(self):
        nx, ny = self.size
        sym = 1 if self.edge is "symmetric" else 0
        spin_count = (ny + sym) * nx + ny * (nx + sym)
        pos = np.zeros((spin_count, 2), dtype=float)
        angle = np.zeros(spin_count, dtype=float)

        labels = []

        a = self.lattice_spacing
        y = 0
        i = 0
        for row in range(0, 2 * ny + sym):
            is_vert = row % 2 # 1 for vert, 0 for horiz
            ncols = nx
            x = 0

            if is_vert:
                # vertical row
                ncols += sym
            else:
                # horizontal row
                x += a/2

            for col in range(0, ncols):
                if is_vert:
                    angle[i] = np.pi/2
                pos[i] = [x, y]

                label = (row, col)
                labels.append(label)

                x += a
                i += 1

            y += a / 2

        self.labels = np.array(labels)

        return pos, angle

    def _init_spin(self, init):
        if isinstance(init, str) and init == 'ground':
            for row, col in self.labels:
                i = self.L[row, col]
                if row % 4 == 0 and col % 2 == 1:
                    self.flip(i)
                elif row % 4 == 2 and col % 2 == 0:
                    self.flip(i)
                elif row % 4 == 1 and col % 2 == 0:
                    self.flip(i)
                elif row % 4 == 3 and col % 2 == 1:
                    self.flip(i)
        else:
            super()._init_spin(init)

    _vertex_size = (3, 3)

    @property
    def _default_cell_size(self):
        return (self.lattice_spacing/2, self.lattice_spacing/2)

class SquareSpinIceOpen(SquareSpinIce):
    def __init__(self, *, neighbor_distance=sqrt(2), **kwargs):
        kwargs['neighbor_distance'] = neighbor_distance

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        angle[:] = np.pi/4

        for i in range(len(pos)):
            row, col = self.labels[i]
            if row % 2 == 0 and col % 2 == 1:
                angle[i] += np.pi/2
            elif row % 2 == 1 and col % 2 == 0:
                angle[i] += np.pi/2

        return pos, angle

    _vertex_size = (2, 2)

class PinwheelSpinIceDiamond(SquareSpinIceClosed):
    def __init__(self, *, spin_angle=45, neighbor_distance=10, **kwargs):
        self.spin_angle = spin_angle
        kwargs['neighbor_distance'] = neighbor_distance

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        # rotate each island by spin_angle
        angle[:] -= np.deg2rad(self.spin_angle)

        return pos, angle

class PinwheelSpinIceLuckyKnot(SquareSpinIceOpen):
    def __init__(self, *, spin_angle=45, neighbor_distance=10*sqrt(2), **kwargs):
        self.spin_angle = spin_angle
        kwargs['neighbor_distance'] = neighbor_distance

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        # rotate each island by spin_angle
        angle[:] -= np.deg2rad(self.spin_angle)

        return pos, angle

class PinwheelSpinIceRandom(PinwheelSpinIceDiamond):
    def __init__(self, *, spin_angle_disorder=0, **kwargs):
        self.spin_angle_disorder = spin_angle_disorder

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        # rotate each island by spin_angle
        std = self.spin_angle_disorder
        angle[:] += np.deg2rad(np.random.normal(0, std, angle.shape))

        return pos, angle

class KagomeSpinIce(SpinIce):
    def _init_geometry(self):
        labels = []

        nx, ny = self.size

        n_top_bot = (2 * nx * 2)
        n_mid = (2 * nx + 1) * (ny - 1)
        n_vert = (nx + 1) * ny
        spin_count = n_top_bot + n_mid + n_vert

        pos = np.zeros((spin_count, 2), dtype=float)
        angle = np.zeros(spin_count, dtype=float)

        n_rows = 2 * ny + 1
        last_row = n_rows - 1

        a = self.lattice_spacing
        y = 0
        i = 0
        for row in range(0, n_rows):
            if row % 2 == 0:
                # "horizontal" magnets (+/-30 degree)
                n_cols = 2 * nx + 1
                if row == 0 or row == last_row:
                    # first and last row has 1 less element
                    n_cols -= 1

                x = a/2
                col0 = 0
                if ny % 2 == 0 and row == last_row:
                    # even number of magnets, skip first magnet
                    x += a
                    col0 += 1

                for col in range(col0, col0 + n_cols):
                    pos[i] = [x, y]
                    angle[i] = np.deg2rad(30)
                    if row % 4 == 0 and col % 2 == 0:
                        angle[i] *= -1
                    if row % 4 == 2 and col % 2 == 1:
                        angle[i] *= -1

                    label = (row, col)
                    labels.append(label)

                    x += a
                    i += 1
            else:
                # vertical magnets (90 degrees)
                n_cols = nx + 1
                x = 0
                if row % 4 == 3:
                    x += a
                for col in range(0, n_cols):
                    pos[i] = [x, y]
                    angle[i] = np.deg2rad(90)

                    label = (row, col)
                    labels.append(label)

                    x += 2 * a
                    i += 1

            y += a * sqrt(3) / 2

        self.labels = np.array(labels)

        return pos, angle

    _vertex_size = (2,3)

    @property
    def _default_cell_size(self):
        sizex = self.lattice_spacing/2
        sizey = sqrt(3)*self.lattice_spacing/2
        return (sizex, sizey)


if __name__ == '__main__':
    # np.random.seed(0xdeadbeef)
    # random.seed(0xdeadbeef)

    hc = 11.25
    size = (8,8)
    # spin_ice = SpinIce(size, hc=hc, disorder=0.0)
    spin_ice = SquareSpinIceClosed(size=size, hc=hc, disorder=0.0)
    # spin_ice = SquareSpinIceOpen(size, hc=hc, disorder=0.0)

    # External field
    H = 1.1 * hc
    theta = np.pi/4
    h_extx = H * np.cos(theta)
    h_exty = H * np.sin(theta)
    h_ext = np.array([h_extx, h_exty])
    spin_ice.set_h_ext(h_ext)

    # Initial config
    spin_ice.randomize()

    print("Spin:")
    print(spin_ice.spin)
    print("Total energy:", spin_ice.total_energy())

    # Relax
    print("Relaxing", end='')
    t0 = time.time()
    steps = 0
    while spin_ice.step():
        steps += 1
        print(".", flush=True, end='')
    # steps = spin_ice.relax()
    dt = timedelta(seconds=time.time() - t0)
    print("{} steps in {}".format(steps, dt))
    print("Total energy:", spin_ice.total_energy())

    fig, axes = plt.subplots(1, 4, figsize=(16,9), subplot_kw={'aspect': 'equal'}, sharex=True, sharey=True)
    ax_spin, ax_energy, ax_vert, ax_mag  = axes.ravel()

    plt.sca(ax_spin)
    plt.title("Spin")
    spin_ice.plot()

    plt.sca(ax_energy)
    plt.title("Energy")
    spin_ice.plot_energy()

    plt.sca(ax_vert)
    plt.title("Vertex types")
    spin_ice.plot_vertices()

    '''
    plt.sca(ax_mag)
    plt.title("Vertex magnetization")
    spin_ice.plot_vertex_mag()
    '''

    plt.show()



#!/usr/bin/env python
import re

from setuptools import setup, find_namespace_packages

with open('README.rst', 'r') as f:
    long_description = f.read()

with open('flatspin/__init__.py', 'r') as f:
    version = re.search(r'__version__ = \'(.*?)\'', f.read()).group(1)

setup(
    name='flatspin',
    version=version,
    url='https://flatspin.gitlab.io/',
    project_urls={
        "Source Code": "https://gitlab.com/flatspin/flatspin",
        "Documentation": "https://flatspin.readthedocs.io/en/latest/",
        "Bug Tracker": "https://gitlab.com/flatspin/flatspin/issues",
    },
    license='GPLv3',
    author='Johannes H. Jensen',
    author_email='johannes.jensen@ntnu.no',
    description='Artificial Spin Ice simulator',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    packages=find_namespace_packages(include=["flatspin", "flatspin.*"]),
    package_data={
        "": ["*.c"],
    },
    python_requires='>=3.5',
    install_requires=[
        'numpy',
        'scipy',
        'pandas>=0.23.0',
        'tables',
        'matplotlib',
        'pyopencl',
        'pillow',
        'joblib',
        'numba',
        'tqdm',
        'scikit-image',
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-benchmark',
            'sphinx',
            'sphinx-autoapi',
            'sphinx_rtd_theme',
        ],
        'docs': [
            'sphinx',
        ],
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': [
            'flatspin-run = flatspin.tools.run:main',
            'flatspin-run-sweep = flatspin.tools.run_sweep:main',
            'flatspin-plot-table = flatspin.tools.plot_table:main',
            'flatspin-plot = flatspin.tools.plot:main',
            'flatspin-vectors = flatspin.tools.vectors:main',
            'flatspin-vertices = flatspin.tools.vertices:main',
            'flatspin-stats = flatspin.tools.stats:main',
            'flatspin-convert = flatspin.tools.convert:main',
            'flatspin-fsck = flatspin.tools.fsck:main',
            'flatspin-inspect = flatspin.tools.inspect:main',
            'flatspin-progress = flatspin.tools.progress:main',
        ],
    }
)
